// Türklingel mit 100 Melodien M. Baudisch 12/2008
// 20170424, de0508, Uwe
// Anpassung an aktuellen avr gcc version 4.9.2 (GCC) auf meinem System
// 
// kombiniert mit dem Waveform Generator von David Johnson-Davies 
// http://www.technoblogy.com/show?QVN
//
// und Wavetable + Huellkurven von ELM ChaN
// http://elm-chan.org/works/mxb/report.html
//
// für AVR ATtiny85 mit High-Speed-PWM
//
// Compiler: avr-gcc (GCC) 9.2.0


/*
 * Pinout
 * ------
 *
 * Nutzung          Funktion PIN         PIN  Funktion   Nutzung
 *                               +--v--+
 *                   /RESET  PB5 |1   8| VCC
 * Verstaerker ein   PB3     PB3 |2   7| PB2  SCK/PCINT2 /Taster + interner Pull-Up
 * Lautsprecher inv. OC1B    PB4 |3   6| PB1  MISO/OC1A  Lautsprecher
 *                           GND |4   5| PB0  MOSI
 *                               +-----+
 *
 * Interrupts
 * ----------
 *  PCINT0
 *  TIMER0, Compare Match A
 *
 * Timer
 * -----
 *  Timer 0,  20 kHz
 *  Timer 1, 250 kHz
 */

////////////////////////////////////////
// includes

#include <inttypes.h>       // uint8_t, etc.
#include <stdbool.h>
#include <stdlib.h>         // itoa, utoa

#include <avr/io.h>         // Ports & Pins
#include <avr/interrupt.h>  // sei
#include <avr/eeprom.h>     // eeprom Zugriff
#include <avr/pgmspace.h>   // flash Zugriff
#include <avr/sleep.h>      // sleep_mode

#include <util/delay.h>     // _delay_ms, _delay_us

#include "c_makros.h"
#include "lied.h"           // Melodien, Offsets und Tonhoehen
#include "sawtooth_table.h" // Saegezahn-Tabelle
#include "envelope_table.h" // Huellkurve


////////////////////////////////////////
// global defines

#define     SPEAKER_BITS    ((1<<PB1)|(1<<PB4))
#define     TASTE           (PB2)
#define     AMP_PIN         (PB3)

#define     AMP_INIT        set_bit( DDRB, AMP_PIN);
#define     AMP_ON          set_bit( PORTB, AMP_PIN);
#define     AMP_OFF         clear_bit( PORTB, AMP_PIN);

#define     CHANNELS        4
#define     ENVELOPE_STEP   4


////////////////////////////////////////
// global types
typedef struct
{
    uint16_t phase_accu;
    uint16_t phase_step;
    uint16_t envelope;
} voice_t;

////////////////////////////////////////
// register variables

register uint16_t soundtime asm("r2");
register int16_t sum asm("r4");

////////////////////////////////////////
// global variables

uint16_t liedcount;
uint16_t liedende;
uint8_t liednr;
uint8_t speed;
uint8_t dauer;
uint8_t hoehe;
uint8_t note;
volatile uint8_t rand_value;

voice_t voice[ CHANNELS];
uint8_t channel = 0;
volatile int16_t code_max = 0;


////////////////////////////////////////
// function prototypes
void PLL_enable( void);
void sound_on( void);
void sound_off( void);


////////////////////////////////////////
// helper functions

void setup( void)
{

    set_sleep_mode( SLEEP_MODE_PWR_DOWN);

    // Verstaerker: Ausgang
    AMP_INIT;
    AMP_OFF;

    // Taster: Eingang + Pull-Up
    clear_bit( DDRB, TASTE);
    set_bit( PORTB, TASTE);

    // Pinchange an PB2 konfigurieren
    set_bit( PCMSK, PCINT2);

    // Pinchange Interrupt generell zulassen
    set_bit( GIMSK, PCIE);

    // Timer Interrupts aus
    TIMSK = 0;       

    // 64 MHz PLL einschalten und als Quelle fuer Timer1
    PLL_enable();

    // Konfiguration Timer/Counter1 fuer PWM-Ausgabe
    // Frequency = 64 MHz / 256 = 250 kHz
    TCCR1 = (1<<PWM1A)|(2<<COM1A0)|(1<<CS10);
    GTCCR = (1<<PWM1B)|(3<<COM1B0);
    OCR1C = 255;

    // Konfiguration Timer/Counter0 fuer 20kHz interrupt, Sampleausgabe
    // SysClk: 16 MHz
    TCCR0A = 3<<WGM00;             // Fast PWM
    TCCR0B = 1<<WGM02 | 2<<CS00;   // 1/8 prescale = 2 MHz
    TIMSK  = 1<<OCIE0A;            // Enable compare match, disable overflow
    OCR0A  = 99;                   // Divide by 100 = 20 kHz

    // da Variablen in Registern, explizite Initialisierung noetig!
    soundtime   = 0;

    // Interrupts global zulassen
    sei();

}


void loop( void)
{
    if( liednr > 0)         // wenn Melodie aktiv
    {
        if( soundtime == 0) // wenn kein Ton oder Pause läuft
        {
            // keine Tonzeitdauer aktiv
            if (!speed)     // wir sind am Liedanfang
            {
                liedcount = eeprom_read_word( &liedofs[ liednr-1]);
                if( liednr == ANZ) 
                {
                    liedende = sizeof( lied);
                }
                else
                {
                    // bei letztem Lied ist Endwert Größe des Arrays
                    // ansonsten 1 Byte vor nächstem Lied
                    liedende = eeprom_read_word( &liedofs[ liednr]); 
                }
                speed = pgm_read_byte( &lied[ liedcount]);
                liedcount++;
            } 
            else                // mitten im Lied
            {
                // Ton ausgeben
                note  = pgm_read_byte( &lied[ liedcount]); //Note holen
                hoehe = note & 0x1f;    // Tonhöhe ausmaskieren (5 Bit)
                dauer = eeprom_read_byte( &zeit[ note >> 5]); // 5 Bit nach rechts schieben und Zeit holen
                if( hoehe > 0)          // wenn keine Pause 
                {
                    voice[ channel].envelope = 0xffff;
                    voice[ channel].phase_step = eeprom_read_word( &ton[ hoehe-1]);
                    // naechster Kanal
                    channel++;
                    if( channel == CHANNELS) channel = 0;
                }


                // Bereich dauer: 2..24
                // Bereich speed: 2..5
                // Bereich soundtime: 4..120
                // da jetzt 20k statt 50 -> zusaetzlich Faktor 400

                // da Variable vom Interrupt während der Zuweisung geaendert werden koennte
                cli();
                soundtime = dauer * speed * 400;
                sei();
                
                liedcount++;

                // Lied zu Ende
                if( liedcount == liedende)
                {
                    speed  = 0;
                    liednr = 0;
      
                    // warten bis alle Kanaele fertig sind
                    for( uint8_t index = 0; index < CHANNELS; index++)
                    {
                        uint16_t temp;
                        do
                        {
                            cli();
                            temp = voice[ index].envelope;
                            sei();
                        } while( temp > 0);
                    }
                }
            }
        } // (!soundtime)
    } // liednr
    else 
    {
        sound_off();
        sleep_mode();
        
        // Start nach Aufwachen erfolgt hier!
        PLL_enable();
        rand_value++;
        rand_value = rand_value % ANZ;
        liednr = rand_value + 1;
        sound_on();
    }
}




// Interruptroutine, fuer Aufwachen aus dem Sleep
ISR( PCINT0_vect)
{
    // Lied weiterzaehlen
    rand_value++;
    rand_value = rand_value % ANZ;
}


// naechstes Sample ausgeben
// Laufzeit @16 MHz 23 us ... 45 us (bei einer Wiederholrate von 50 us)
// die ISR sorgt fuer eine CPU-Auslastung von 46% bis 90%
ISR( TIMER0_COMPA_vect) 
{
    uint8_t index;
    uint8_t wavetable_index;
    uint8_t envtable_index;

    // nur hi-Byte verwenden
    sum = sum >> 8;

    // berechnete Werte ausgeben, um Jitter zu reduzieren
    OCR1A = (uint8_t)sum;
    OCR1B = (uint8_t)sum;

    // Berechnung zuruecksetzen
    sum = 0;

    // alle Kanaele durchgehen
    for( index = 0; index < CHANNELS; index++)
    {
        // Phasenakkumulator weitersetzen
        voice[ index].phase_accu += voice[ index].phase_step;
        // Eintrag aus Wellenform-Tabelle (hi-Byte vom Phasenakku)
        wavetable_index = voice[ index].phase_accu >> 8;
        // Eintrag aus Huellkurventabelle (auch hi-Byte)
        envtable_index = voice[ index].envelope >> 8;
        // beides multplizieren und zur Summe addieren
        sum += (int8_t)pgm_read_byte( &sawtooth_table[ wavetable_index]) * pgm_read_byte( &envelope_table[ envtable_index]);
        // Huellkurve weitersetzen (die Huellkurve wird rueckwaerts ausgelesen)
        if( voice[ index].envelope > ENVELOPE_STEP)
        {
            voice[ index].envelope -= ENVELOPE_STEP;
        }
        else
        {
            voice[ index].envelope = 0;
        }
    }
    
    // DC-Offset addieren
    sum += (128 << 8);


    if( soundtime > 0)
    {
        soundtime--;
    }
}


void PLL_enable( void)
{
    // enable PLL
    set_bit( PLLCSR, PLLE);
    _delay_us( 100);
    // wait for lock in
    while(!(PLLCSR & (1<<PLOCK)));
    _delay_us( 100);
    // enable timer 1 async mode
    PLLCSR |= (1<<PCKE)|(1<<PLLE);
}


void sound_on( void)
{
    // Ausgaenge aktivieren
    DDRB |= SPEAKER_BITS;

    // Verstaerker anschalten
    AMP_ON;

    // PWM hochfahren
    uint8_t value = 0;
    do
    {
        OCR1A = value;
        value++;
        _delay_us( 20);
    } while( value < 128);

    // Restart 20 kHz Timer Interrupts
    set_bit( TIMSK, OCIE0A);
}


void sound_off( void)
{
    // Stop 20 kHz Timer Interrupts
    clear_bit( TIMSK, OCIE0A);

    // Ausgang langsam runterfahren
    uint8_t value = 128;
    do
    {
        OCR1A = value;
        value--;
        _delay_ms( 1);
    }
    while( value > 0);

    // Verstaerker abschalten
    AMP_OFF;

    // Ausgaenge hochohmig schalten
    DDRB &= ~(SPEAKER_BITS);
}



int main( void)
{
    setup();

    while( 1)
    {
        loop();
    }
}
